#
# Copyright (c) 2015-2022 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_auth_ldap.utils module

This module defines a small set of utility functions.
"""

__docformat__ = 'restructuredtext'


def get_single_value(entry):
    """Get single value for provided input"""
    if isinstance(entry, (list, tuple)):
        return '<br />'.join(map(get_single_value, entry))
    if isinstance(entry, bytes):
        return entry.decode('utf-8', 'ignore')
    return str(entry)


def get_dict_values(entries: dict, attributes: list):
    """Get given attributes from entries"""
    values = {}
    for attr in attributes:
        values[attr] = get_single_value(entries[attr])
    return values


def get_formatted_value(formatter: str, **entries: dict):
    """Get formatted value of given formatter with provided entries"""
    for key, value in entries.items():
        entries[key] = get_single_value(value)
    return formatter.format(**entries)



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_auth_ldap.zmi module

This module defines view and content providers used to manage LDAP directory plug-ins.
"""

import base64

from ldap3 import ALL_ATTRIBUTES, BASE
from pyramid.decorator import reify
from zope.interface import Interface, implementer

from pyams_auth_ldap.interfaces import ILDAPPlugin, LDAP_PLUGIN_LABEL
from pyams_auth_ldap.query import LDAPQuery
from pyams_auth_ldap.utils import get_single_value
from pyams_auth_ldap.zmi.interfaces import ILDAPPluginConnectionSubform, \
    ILDAPPluginGroupsSchemaSubform, ILDAPPluginSearchSettingsSubform, \
    ILDAPPluginUsersSchemaSubform
from pyams_form.ajax import ajax_form_config
from pyams_form.browser.checkbox import SingleCheckBoxFieldWidget
from pyams_form.field import Fields
from pyams_form.interfaces.form import IFormFields, IInnerTabForm
from pyams_form.subform import InnerAddForm, InnerEditForm
from pyams_layer.interfaces import IPyAMSLayer
from pyams_pagelet.pagelet import pagelet_config
from pyams_security.interfaces import IPlugin, ISecurityManager
from pyams_security.interfaces.base import MANAGE_SECURITY_PERMISSION
from pyams_security_views.zmi import SecurityPluginsTable
from pyams_security_views.zmi.plugin import SecurityPluginAddForm, \
    SecurityPluginAddMenu, SecurityPluginPropertiesEditForm
from pyams_skin.interfaces.viewlet import IContentSuffixViewletManager
from pyams_table.column import GetAttrColumn
from pyams_table.interfaces import IColumn, IValues
from pyams_utils.adapter import ContextRequestViewAdapter, adapter_config
from pyams_utils.registry import get_utility
from pyams_utils.url import absolute_url
from pyams_viewlet.viewlet import viewlet_config
from pyams_zmi.form import AdminModalDisplayForm
from pyams_zmi.interfaces import IAdminLayer, TITLE_SPAN_BREAK
from pyams_zmi.interfaces.form import IFormTitle
from pyams_zmi.interfaces.table import IInnerTable
from pyams_zmi.interfaces.viewlet import IContextAddingsViewletManager
from pyams_zmi.search import SearchForm, SearchResultsView, SearchView
from pyams_zmi.table import I18nColumnMixin, Table
from pyams_zmi.utils import get_object_label


__docformat__ = 'restructuredtext'

from pyams_auth_ldap import _  # pylint: disable=ungrouped-imports


#
# LDAP plug-in add form
#


@viewlet_config(name='add-ldap-plugin.menu',
                context=ISecurityManager, layer=IAdminLayer, view=SecurityPluginsTable,
                manager=IContextAddingsViewletManager, weight=40,
                permission=MANAGE_SECURITY_PERMISSION)
class LDAPPluginAddMenu(SecurityPluginAddMenu):
    """LDAP plug-in add menu"""

    label = _("Add LDAP directory...")
    href = 'add-ldap-plugin.html'


@ajax_form_config(name='add-ldap-plugin.html',
                  context=ISecurityManager, layer=IPyAMSLayer,
                  permission=MANAGE_SECURITY_PERMISSION)
class LDAPPluginAddForm(SecurityPluginAddForm):
    """LDAP plug-in add form"""

    content_factory = ILDAPPlugin
    content_label = LDAP_PLUGIN_LABEL

    fields = Fields(IPlugin).omit('__parent__', '__name__')
    fields['enabled'].widget_factory = SingleCheckBoxFieldWidget


@adapter_config(name='connection',
                required=(ISecurityManager, IAdminLayer, LDAPPluginAddForm),
                provides=IInnerTabForm)
@implementer(ILDAPPluginConnectionSubform)
class LDAPPluginConnectionAddForm(InnerAddForm):
    """LDAP plug-in add form connection sub-form"""

    title = _("Connection")

    weight = 1


@adapter_config(name='users-schema',
                required=(ISecurityManager, IAdminLayer, LDAPPluginAddForm),
                provides=IInnerTabForm)
class LDAPPluginUsersSchemaAddForm(InnerAddForm):
    """LDAP plug-in add form users schema properties sub-form"""

    title = _("Users schema")

    fields = Fields(ILDAPPlugin).select('base_dn', 'search_scope', 'login_attribute',
                                        'login_query', 'uid_attribute', 'uid_query',
                                        'title_format', 'mail_attribute',
                                        'user_extra_attributes')
    weight = 10


@adapter_config(name='groups-schema',
                required=(ISecurityManager, IAdminLayer, LDAPPluginAddForm),
                provides=IInnerTabForm)
class LDAPPluginGroupsSchemaAddForm(InnerAddForm):
    """LDAP plug-in add form groups schema properties sub-form"""

    title = _("Groups schema")

    fields = Fields(ILDAPPlugin).select('groups_base_dn', 'groups_search_scope', 'group_prefix',
                                        'group_uid_attribute', 'group_title_format',
                                        'group_members_query_mode', 'groups_query',
                                        'group_members_attribute', 'user_groups_attribute',
                                        'group_mail_mode', 'group_replace_expression',
                                        'group_mail_attribute', 'group_extra_attributes')
    weight = 20


@adapter_config(name='search-settings',
                required=(ISecurityManager, IAdminLayer, LDAPPluginAddForm),
                provides=IInnerTabForm)
class LDAPPluginSearchSettingsAddForm(InnerAddForm):
    """LDAP plug-in add form search settings sub-form"""

    title = _("Search settings")

    fields = Fields(ILDAPPlugin).select('users_select_query', 'users_search_query',
                                        'groups_select_query', 'groups_search_query')
    weight = 30


#
# LDAP plug-in properties edit form
#

@ajax_form_config(name='properties.html', context=ILDAPPlugin, layer=IPyAMSLayer)
class LDAPPluginPropertiesEditForm(SecurityPluginPropertiesEditForm):
    """LDAP plug-in properties edit form"""


@adapter_config(name='connection',
                required=(ILDAPPlugin, IAdminLayer, LDAPPluginPropertiesEditForm),
                provides=IInnerTabForm)
@implementer(ILDAPPluginConnectionSubform)
class LDAPPluginConnectionEditForm(InnerEditForm):
    """LDAP plug-in connection edit form"""

    title = _("Connection")

    weight = 1


@adapter_config(name='users-schema',
                required=(ILDAPPlugin, IAdminLayer, LDAPPluginPropertiesEditForm),
                provides=IInnerTabForm)
class LDAPPluginUsersSchemaEditForm(InnerEditForm):
    """LDAP plug-in users schema properties edit form"""

    title = _("Users schema")

    fields = Fields(ILDAPPlugin).select('base_dn', 'search_scope', 'login_attribute',
                                        'login_query', 'uid_attribute', 'uid_query',
                                        'title_format', 'mail_attribute',
                                        'user_extra_attributes')
    weight = 10


@adapter_config(name='groups-schema',
                required=(ILDAPPlugin, IAdminLayer, LDAPPluginPropertiesEditForm),
                provides=IInnerTabForm)
class LDAPPluginGroupsSchemaEditForm(InnerEditForm):
    """LDAP plug-in groups schema properties edit form"""

    title = _("Groups schema")

    fields = Fields(ILDAPPlugin).select('groups_base_dn', 'groups_search_scope', 'group_prefix',
                                        'group_uid_attribute', 'group_title_format',
                                        'group_members_query_mode', 'groups_query',
                                        'group_members_attribute', 'user_groups_attribute',
                                        'group_mail_mode', 'group_replace_expression',
                                        'group_mail_attribute', 'group_extra_attributes')
    weight = 20


@adapter_config(name='search-settings',
                required=(ILDAPPlugin, IAdminLayer, LDAPPluginPropertiesEditForm),
                provides=IInnerTabForm)
class LDAPPluginSearchSettingsEditForm(InnerEditForm):
    """LDAP plug-in search settings edit form"""

    title = _("Search settings")

    fields = Fields(ILDAPPlugin).select('users_select_query', 'users_search_query',
                                        'groups_select_query', 'groups_search_query')
    weight = 30


#
# LDAP plug-in sub-forms fields adapters
#

@adapter_config(required=(Interface, IAdminLayer, ILDAPPluginConnectionSubform),
                provides=IFormFields)
def ldap_plugin_connection_fields(context, request, view):  # pylint: disable=unused-argument
    """LDAP plugin connection fields"""
    return Fields(ILDAPPlugin).select('server_uri', 'start_tls',
                                      'bind_dn', 'bind_password', 'bind_mode')


@adapter_config(required=(Interface, IAdminLayer, ILDAPPluginUsersSchemaSubform),
                provides=IFormFields)
def ldap_plugin_users_schema_fields(context, request, view):  # pylint: disable=unused-argument
    """LDAP plugin users schema fields"""
    return Fields(ILDAPPlugin).select('base_dn', 'search_scope', 'login_attribute',
                                      'login_query', 'uid_attribute', 'uid_query',
                                      'title_format', 'mail_attribute',
                                      'user_extra_attributes')


@adapter_config(required=(Interface, IAdminLayer, ILDAPPluginGroupsSchemaSubform),
                provides=IFormFields)
def ldap_plugin_groups_schema_fields(context, request, view):  # pylint: disable=unused-argument
    """LDAP plugin groups schema fields"""
    return Fields(ILDAPPlugin).select('groups_base_dn', 'groups_search_scope', 'group_prefix',
                                      'group_uid_attribute', 'group_title_format',
                                      'group_members_query_mode', 'groups_query',
                                      'group_members_attribute', 'user_groups_attribute',
                                      'group_mail_mode', 'group_replace_expression',
                                      'group_mail_attribute', 'group_extra_attributes')


@adapter_config(required=(Interface, IAdminLayer, ILDAPPluginSearchSettingsSubform),
                provides=IFormFields)
def ldap_plugin_search_settings_fields(context, request, view):  # pylint: disable=unused-argument
    """LDAP plugin search settings fields"""
    return Fields(ILDAPPlugin).select('users_select_query', 'users_search_query',
                                      'groups_select_query', 'groups_search_query')

#
# LDAP folder search view
#

class LDAPPluginSearchForm(SearchForm):  # pylint: disable=abstract-method
    """LDAP plug-in search form"""

    title = _("LDAP directory search form")

    @property
    def back_url(self):
        """Form back URL getter"""
        manager = get_utility(ISecurityManager)
        return absolute_url(manager, self.request, 'security-plugins.html')


@pagelet_config(name='search.html', context=ILDAPPlugin, layer=IPyAMSLayer,
                permission=MANAGE_SECURITY_PERMISSION)
class LDAPPluginSearchView(SearchView):
    """LDAP plug-in search view"""

    title = _("LDAP directory search form")
    search_form = LDAPPluginSearchForm


class LDAPPluginSearchResultsTable(Table):
    """LDAP plug-in search results table"""

    @reify
    def data_attributes(self):
        attributes = super().data_attributes
        attributes.setdefault('tr', {}).update({
            'data-ams-url': lambda row, col: absolute_url(self.context, self.request,
                                                          'ldap-properties.html',
                                                          query={'dn': row[0]}),
            'data-toggle': 'modal'
        })
        return attributes

    batch_size = 999


@adapter_config(required=(ILDAPPlugin, IAdminLayer, LDAPPluginSearchResultsTable),
                provides=IValues)
class LDAPPluginSearchResultsValues(ContextRequestViewAdapter):
    """LDAP plug-in search results values"""

    @property
    def values(self):
        """LDAP plug-in search table results getter"""
        yield from self.context.get_search_results({
            'query': self.request.params.get('form.widgets.query')
        })


class LDAPColumn(I18nColumnMixin, GetAttrColumn):
    """Base LDAP column"""

    def get_value(self, obj):
        value = obj[1].get(self.attr_name, ())
        return '<br /> '.join(value) if isinstance(value, (list, tuple)) else value


@adapter_config(name='uid',
                required=(ILDAPPlugin, IAdminLayer, LDAPPluginSearchResultsTable),
                provides=IColumn)
class LDAPPluginSearchUIDColumn(LDAPColumn):
    """LDAP plug-in search UID column"""

    i18n_header = _("UID")

    @property
    def attr_name(self):
        return self.context.uid_attribute

    weight = 10


@adapter_config(name='cn',
                required=(ILDAPPlugin, IAdminLayer, LDAPPluginSearchResultsTable),
                provides=IColumn)
class LDAPPluginSearchCNColumn(LDAPColumn):
    """LDAP plug-in search CN column"""

    i18n_header = _("Common name")
    attr_name = 'cn'

    weight = 20


@adapter_config(name='mail',
                required=(ILDAPPlugin, IAdminLayer, LDAPPluginSearchResultsTable),
                provides=IColumn)
class LDAPPluginSearchMailColumn(LDAPColumn):
    """LDAP plug-in search mail column"""

    i18n_header = _("Mail address")
    attr_name = 'mail'

    weight = 30


@pagelet_config(name='search-results.html', context=ILDAPPlugin, layer=IPyAMSLayer,
                permission=MANAGE_SECURITY_PERMISSION, xhr=True)
class LDAPPluginSearchResultsView(SearchResultsView):
    """LDAP plug-in search results view"""

    table_label = _("Search results")
    table_class = LDAPPluginSearchResultsTable


#
# LDAP entry display view
#

@pagelet_config(name='ldap-properties.html',
                context=ILDAPPlugin, layer=IPyAMSLayer,
                permission=MANAGE_SECURITY_PERMISSION)
class LDAPEntryPropertiesDisplayForm(AdminModalDisplayForm):
    """LDAP entry properties display form"""

    modal_class = 'modal-xl'

    legend = _("LDAP entry attributes")
    fields = Fields(Interface)

    @reify
    def ldap_entry(self):
        """LDAP entry getter"""
        plugin = self.context
        conn = plugin.get_connection()
        dn = self.request.params.get('dn')  # pylint: disable=invalid-name
        query = LDAPQuery(dn, '(objectclass=*)', BASE, ALL_ATTRIBUTES)
        result = query.execute(conn)
        if not result or len(result) > 1:
            return {}
        dn, attributes = result[0]  # pylint: disable=invalid-name
        photo = attributes.get('jpegPhoto')
        if photo is not None:
            if isinstance(photo, list):
                photo = photo[0]
            attributes['jpegPhoto'] = [
                '<img src="data:image/jpeg;base64,{0}" />'.format(
                    base64.encodebytes(photo).decode())
            ]
        photo = attributes.get('thumbnailPhoto')
        if photo is not None:
            if isinstance(photo, list):
                photo = photo[0]
            attributes['thumbnailPhoto'] = [
                '<img src="data:image/jpeg;base64,{0}" />'.format(
                    base64.encodebytes(photo).decode())
            ]
        result = sorted(attributes.items(), key=lambda x: x[0])
        return {
            'dn': dn,
            'attributes': result
        }


@adapter_config(required=(ILDAPPlugin, IAdminLayer, LDAPEntryPropertiesDisplayForm),
                provides=IFormTitle)
def ldap_entry_display_form_title(context, request, form):
    """LDAP entry display form title getter"""
    translate = request.localizer.translate
    return TITLE_SPAN_BREAK.format(
        get_object_label(context, request, form),
        translate(_("DN: {}")).format(form.ldap_entry.get('dn', _('unknown'))))


@viewlet_config(name='entry-properties',
                context=ILDAPPlugin, layer=IAdminLayer,
                view=LDAPEntryPropertiesDisplayForm,
                manager=IContentSuffixViewletManager)
@implementer(IInnerTable)
class LDAPEntryPropertiesTable(Table):
    """LDAP entry properties table"""

    def __init__(self, context, request, view, manager):
        super().__init__(context, request)
        self.view = view
        self.manager = manager

    batch_size = 999


@adapter_config(required=(ILDAPPlugin, IAdminLayer, LDAPEntryPropertiesTable),
                provides=IValues)
class LDAPEntryPropertiesTableValues(ContextRequestViewAdapter):
    """LDAP entry properties table values"""

    @property
    def values(self):
        """LDAP entry attributes getter"""
        yield from self.view.view.ldap_entry.get('attributes', ())


@adapter_config(name='attribute',
                required=(ILDAPPlugin, IAdminLayer, LDAPEntryPropertiesTable),
                provides=IColumn)
class LDAPEntryAttributeColumn(I18nColumnMixin, GetAttrColumn):
    """LDAP entry attribute column"""

    i18n_header = _("Attribute")
    weight = 10

    def get_value(self, obj):
        return obj[0]


@adapter_config(name='values',
                required=(ILDAPPlugin, IAdminLayer, LDAPEntryPropertiesTable),
                provides=IColumn)
class LDAPEntryValuesColumn(I18nColumnMixin, GetAttrColumn):
    """LDAP entry values column"""

    i18n_header = _("Value")
    weight = 20

    def get_value(self, obj):
        """Value getter"""
        return get_single_value(obj[1])



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_auth_ldap.interfaces module

"""

import re

from ldap3 import BASE, LEVEL, SUBTREE, ANONYMOUS, SIMPLE, SASL, NTLM
from zope.interface import Attribute, Interface
from zope.schema import Bool, Choice, TextLine
from zope.schema.vocabulary import SimpleTerm, SimpleVocabulary

from pyams_security.interfaces.plugin import IAuthenticationPlugin, IDirectorySearchPlugin

from pyams_auth_ldap import _


#
# Search scopes vocabulary
#


SEARCH_SCOPES = {
    BASE: _("Base object"),
    LEVEL: _("Single level"),
    SUBTREE: _("Whole subtree")
}

SEARCH_SCOPES_VOCABULARY = SimpleVocabulary([
    SimpleTerm(v, title=t) for v, t in SEARCH_SCOPES.items()
])


#
# Group members query mode
#

QUERY_MEMBERS_FROM_GROUP = 'group'
QUERY_MEMBERS_FROM_MEMBER = 'member'

GROUP_MEMBERS_QUERY_MODE = {
    QUERY_MEMBERS_FROM_GROUP: _("Use group attribute to get members list"),
    QUERY_MEMBERS_FROM_MEMBER: _("Use member attribute to get groups list")
}

GROUP_MEMBERS_QUERY_MODE_VOCABULARY = SimpleVocabulary([
    SimpleTerm(v, title=t) for v, t in GROUP_MEMBERS_QUERY_MODE.items()
])


#
# Group mail mode
#

NO_GROUP_MAIL_MODE = 'none'
INTERNAL_GROUP_MAIL_MODE = 'internal'
REDIRECT_GROUP_MAIL_MODE = 'redirect'

GROUP_MAIL_MODE = {
    NO_GROUP_MAIL_MODE: _("none (only use members own mail address)"),
    INTERNAL_GROUP_MAIL_MODE: _("Use group internal attribute"),
    REDIRECT_GROUP_MAIL_MODE: _("Use another group internal attribute")
}

GROUP_MAIL_MODE_VOCABULARY = SimpleVocabulary([
    SimpleTerm(v, title=t) for v, t in GROUP_MAIL_MODE.items()
])


class ILDAPBaseInfo(Interface):
    """LDAP base entry interface"""

    dn = Attribute("LDAP DN")

    attributes = Attribute("Entry attributes")


class ILDAPUserInfo(ILDAPBaseInfo):
    """LDAP user entry interface"""


class ILDAPGroupInfo(ILDAPBaseInfo):
    """LDAP group entry interface"""

    def get_members(self):
        """Get all group members"""


DEFAULT_UID_QUERY = '(uid={login})'
DEFAULT_USER_SEARCH = '(|(givenName={query}*)(sn={query}*))'
DEFAULT_GROUP_SEARCH = '(cn=*{query}*)'


LDAP_PLUGIN_LABEL = _("LDAP directory plug-in")


class ILDAPPlugin(IAuthenticationPlugin, IDirectorySearchPlugin):
    """LDAP authentication plug-in interface"""

    server_uri = TextLine(title=_("LDAP server URI"),
                          description=_("Full URI (including protocol) of LDAP server"),
                          default="ldap://localhost:389",
                          required=True)

    start_tls = Bool(title=_("Start TLS"),
                     description=_("Should we start an encrypted connection using StartTLS?"),
                     required=True,
                     default=False)

    bind_dn = TextLine(title=_("Bind DN"),
                       description=_("DN used for LDAP bind; keep empty for anonymous"),
                       required=False)

    bind_password = TextLine(title=_("Bind password"),
                             description=_("Password used for LDAP bind"),
                             required=False)

    bind_mode = Choice(title=_("Bind mode"),
                       description=_("Authentication mode used for binding"),
                       values=(ANONYMOUS, SIMPLE, SASL, NTLM),
                       default=ANONYMOUS,
                       required=False)

    base_dn = TextLine(title=_("Base DN"),
                       description=_("LDAP base DN"),
                       required=True)

    search_scope = Choice(title=_("Search scope"),
                          vocabulary=SEARCH_SCOPES_VOCABULARY,
                          default=SUBTREE,
                          required=True)

    login_attribute = TextLine(title=_("Login attribute"),
                               description=_("LDAP attribute used as user login"),
                               required=True,
                               default='uid')

    login_query = TextLine(title=_("Login query"),
                           description=_("Query template used to authenticate user "
                                         "(you can replace login attribute with '{login}')"),
                           required=True,
                           default=DEFAULT_UID_QUERY)

    uid_attribute = TextLine(title=_("UID attribute"),
                             description=_("LDAP attribute used as principal identifier"),
                             required=True,
                             default='dn')

    uid_query = TextLine(title=_("UID query"),
                         description=_("Query template used to get principal information "
                                       "(you can replace UID attribute with '{login}')"),
                         required=True,
                         default=DEFAULT_UID_QUERY)

    title_format = TextLine(title=_("Title format"),
                            description=_("Principal's title format string"),
                            required=True,
                            default='{givenName} {sn}')

    mail_attribute = TextLine(title=_("Mail attribute"),
                              description=_("LDAP attribute storing mail address"),
                              required=True,
                              default='mail')

    user_extra_attributes = TextLine(title=_("Extra attributes"),
                                     description=_("Comma separated list of additional "
                                                   "attributes which will be extracted by "
                                                   "principals"),
                                     required=False)

    groups_base_dn = TextLine(title=_("Groups base DN"),
                              description=_("Base DN used to search LDAP groups; keep empty to "
                                            "disable groups usage"),
                              required=False)

    groups_search_scope = Choice(title=_("Groups search scope"),
                                 vocabulary=SEARCH_SCOPES_VOCABULARY,
                                 default=SUBTREE,
                                 required=False)

    group_prefix = TextLine(title=_("Group prefix"),
                            description=_("Prefix used to identify groups as principals"),
                            required=False,
                            default='group')

    group_uid_attribute = TextLine(title=_("Group UID attribute"),
                                   description=_("LDAP attribute used as group identifier"),
                                   required=False,
                                   default='dn')

    group_title_format = TextLine(title=_("Group title format"),
                                  description=_("Principal's title format string"),
                                  required=True,
                                  default='{cn}')

    group_members_query_mode = Choice(title=_("Members query mode"),
                                      description=_("Define how groups members are defined"),
                                      vocabulary=GROUP_MEMBERS_QUERY_MODE_VOCABULARY,
                                      default='group')

    groups_query = TextLine(title=_("Groups query"),
                            description=_("When members are stored inside a group attribute, "
                                          "this query template is used to get principal groups "
                                          "(based on DN and UID attributes called '{dn}' "
                                          "and '{login}')"),
                            required=False,
                            default="(&(objectClass=groupOfUniqueNames)(uniqueMember={dn}))")

    group_members_attribute = TextLine(title=_("Group members attribute"),
                                       description=_("When groups members are stored inside a "
                                                     "group attribute, this is the attribute "
                                                     "name"),
                                       required=False,
                                       default='uniqueMember')

    user_groups_attribute = TextLine(title=_("User groups attribute"),
                                     description=_("When user groups are stored inside a user "
                                                   "attribute, this is the attribute name"),
                                     required=False,
                                     default='memberOf')

    group_mail_mode = Choice(title=_("Group mail mode"),
                             description=_("Define how an email can be sent to group members"),
                             vocabulary=GROUP_MAIL_MODE_VOCABULARY,
                             required=True,
                             default='none')

    group_replace_expression = TextLine(title=_("DN replace expression"),
                                        description=_("In 'redirect' mail mode, specify source "
                                                      "and target DN parts, separated by a pipe"),
                                        constraint=re.compile(
                                            "[a-zA-Z0-9-=,]+|[a-zA-Z0-9-=,]+").match,
                                        default='ou=access|ou=lists')

    group_mail_attribute = TextLine(title=_("Mail attribute"),
                                    description=_("In 'internal' mail mode, specify name of "
                                                  "group mail attribute"),
                                    required=True,
                                    default='mail')

    group_extra_attributes = TextLine(title=_("Extra attributes"),
                                      description=_("Comma separated list of additional "
                                                    "attributes which will be extracted by "
                                                    "groups principals"),
                                      required=False)

    users_select_query = TextLine(title=_("Users select query"),
                                  description=_("Query template used to select users"),
                                  required=True,
                                  default=DEFAULT_USER_SEARCH)

    users_search_query = TextLine(title=_("Users search query"),
                                  description=_("Query template used to search users"),
                                  required=True,
                                  default=DEFAULT_USER_SEARCH)

    groups_select_query = TextLine(title=_("Groups select query"),
                                   description=_("Query template used to select groups"),
                                   required=True,
                                   default=DEFAULT_GROUP_SEARCH)

    groups_search_query = TextLine(title=_("Groups search query"),
                                   description=_("Query template used to search groups"),
                                   required=True,
                                   default=DEFAULT_GROUP_SEARCH)

    def get_connection(self):
        """Get LDAP connection"""



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_auth_ldap.query module

This module provides a simple wrapper to handle LDAP queries.

"""

import logging


__docformat__ = 'restructuredtext'

LOGGER = logging.getLogger('PyAMS (ldap)')


class LDAPQuery:
    """Object representing an LDAP query"""

    def __init__(self, base_dn, filter_tmpl, scope, attributes):
        self.base_dn = base_dn.strip()
        self.filter_tmpl = filter_tmpl
        self.scope = scope
        self.attributes = attributes

    def __str__(self):
        return ('base_dn={base_dn}, filter_tmpl={filter_tmpl}, '
                'scope={scope}, attributes={attributes}'.format(**self.__dict__))

    def execute(self, conn, **kwargs):
        """Execute an LDAP query"""
        key = (self.base_dn.format(**kwargs), self.filter_tmpl.format(**kwargs))
        LOGGER.debug(f">>> LDAP query: {self.filter_tmpl} (base {self.base_dn})")
        LOGGER.debug(f"  >      args: {kwargs}")
        ret = conn.search(search_scope=self.scope,
                          attributes=self.attributes,
                          *key)
        result, status, request = conn.get_response(ret, get_request=True)
        LOGGER.debug(f"  > LDAP request: {request}")
        LOGGER.debug(f"  < LDAP status: {status}")
        if result is None:
            result = []
        else:
            result = [
                (r['dn'], r['attributes'])
                for r in result
                if 'dn' in r
            ]
        LOGGER.debug(f"<<< LDAP result: {result}")
        return result
